import os
import yaml
from pkg_resources import iter_entry_points
from qtpy import uic

from qtpy.QtCore import Qt, Slot
from qtpy.QtWidgets import QFileDialog, QApplication, QDialog, QTreeWidgetItem, \
    QStyleFactory

from qtpyvcp import TOP_DIR

CHOOSER_DIR = os.path.abspath(os.path.dirname(__file__))

CUSTOM_VCP_DIR = os.path.expanduser('~/linuxcnc/vcps')
EXAMPLE_VCP_DIR = os.path.join(TOP_DIR, 'examples')

class VCPChooser(QDialog):
    def __init__(self, opts):
        super(VCPChooser, self).__init__()
        uic.loadUi(os.path.join(CHOOSER_DIR, 'vcp_chooser.ui'), self)

        self.setAttribute(Qt.WA_DeleteOnClose, True)

        self.opts = opts
        self._vcp_data = {}

        self.selection = self.vcpTreeView.selectionModel()

        # example VCP section
        category = QTreeWidgetItem(self.vcpTreeView)
        category.setText(0, 'Example VCPs')
        category.setFlags(Qt.ItemIsEnabled)

        # add example VCPs to the treeview
        for entry_point in iter_entry_points(group='qtpyvcp.example_vcp'):
            child = QTreeWidgetItem(category)
            child.setText(0, self.get_vcp_data(entry_point))

        # test VCP section
        category = QTreeWidgetItem(self.vcpTreeView)
        category.setText(0, 'Video Test VCPs')
        category.setFlags(Qt.ItemIsEnabled)

        # add example VCPs to the treeview
        for entry_point in iter_entry_points(group='qtpyvcp.test_vcp'):
            child = QTreeWidgetItem(category)
            child.setText(0, self.get_vcp_data(entry_point))


        # installed VCP section
        category = QTreeWidgetItem(self.vcpTreeView)
        category.setText(0, 'Installed VCPs')
        category.setFlags(Qt.ItemIsEnabled)
        category.setHidden(True)

        # add installed VCPs to the treeview
        for entry_point in iter_entry_points(group='qtpyvcp.vcp'):
            child = QTreeWidgetItem(category)
            child.setText(0, self.get_vcp_data(entry_point))
            category.setHidden(False)

        if os.path.exists(CUSTOM_VCP_DIR):
            category = QTreeWidgetItem(self.vcpTreeView)
            category.setText(0, 'Custom VCPs')
            category.setFlags(Qt.ItemIsEnabled)
            for dir_name in os.listdir(CUSTOM_VCP_DIR):
                if not os.path.isdir(os.path.join(CUSTOM_VCP_DIR, dir_name)):
                    continue
                child = QTreeWidgetItem(category)
                child.setText(0, dir_name)

        self.vcpTreeView.expandAll()
        self.vcpTreeView.activated.connect(self.on_launchVCPButton_clicked)
        self.selection.selectionChanged.connect(self.on_selection_changed)

    def get_vcp_data(self, entry_point):
        vcp_data = {'entry_point_name': entry_point.name}
        vcp = entry_point.load()
        vcp_config_file = vcp.VCP_CONFIG_FILE

        if os.path.exists(vcp_config_file):
            with open(vcp_config_file, 'r') as fh:
                lines = fh.readlines()

            clean = []
            for line in lines:
                # clean jinja templating commands
                if line.strip().startswith('{%') or '{{' in line:
                    continue
                clean.append(line)

            config = yaml.load(''.join(clean), Loader=yaml.SafeLoader)
            if config is not None:
                vcp_data.update(config.get('vcp', {}))
            vcp_name = vcp_data.get('name', entry_point.name)
        else:
            vcp_name = entry_point.name

        self._vcp_data[vcp_name] = vcp_data

        return vcp_name

    def on_selection_changed(self):
        selection = self.selection.selectedRows()
        if selection == []:
            return

        vcp_data = self._vcp_data[selection[0].data()]

        desc = vcp_data.get('description', '')
        self.vcpDescription.setText(desc)


    @Slot()
    def on_launchVCPButton_clicked(self):
        selection = self.selection.selectedRows()
        if selection == []:
            return
        vcp_name = selection[0].data()
        vcp = self._vcp_data[vcp_name].get('entry_point_name', '')

        self.opts.vcp = vcp
        self.accept()

    @Slot()
    def on_cancelButton_clicked(self):
        self.reject()

    @Slot()
    def on_fileButton_pressed(self):
        vcp_file = QFileDialog.getOpenFileName(self,
                              caption="Select VCP File",
                              directory=EXAMPLE_VCP_DIR,
                              filter='VCP Files (*.yml *.yaml *.ui);; All files (*)',
                              options=QFileDialog.DontUseNativeDialog)[0]

        if vcp_file != '':
            self.opts.vcp = vcp_file
            self.accept()

if __name__ == '__main__':
    import sys
    app = QApplication(sys.argv)
    app.setStyle(QStyleFactory.create('Windows'))
    launcher = VCPChooser()
    sys.exit(app.exec_())



import os
from qtpy.QtWidgets import QMenu, QAction

from qtpyvcp import actions
from qtpyvcp.plugins import getPlugin
from qtpyvcp.widgets.dialogs import showDialog

class RecentFilesMenu(QMenu):
    """Recent Files Menu

    Recent files menu provider.

    Args:
        parent (QWidget, optional) : The menus parent. Default to None.
        files (list, optional) : List of initial files in the menu. Defaults to None.
        max_files (int, optional) : Max number of files to show. Defaults to 10.

    Example:

        YAML config::

            main_window:
                menu:
                  - title: File
                    items:
                      - title: &Recent Files
                        provider: qtpyvcp.widgets.recent_files_menu:RecentFilesMenu
                        kwargs:  # optional keyword arguments to pass to the constructor
                          max_files: 15
    """
    def __init__(self, parent=None, files=None, max_files=10):
        super(RecentFilesMenu, self).__init__(parent)

        self._actions = []

        self.status = getPlugin('status')

        for i in range(max_files):
            action = QAction(parent=self,
                             visible=False,
                             triggered=lambda: actions.program.load(self.sender().data()),
                             )

            self._actions.append(action)
            self.addAction(action)

        self.addSeparator()

        action = QAction(parent=self,
                         text='Browse for files ...',
                         triggered=lambda: showDialog('open_file'),
                         )

        self.addAction(action)

        self.update(files or self.status.recent_files())
        self.status.recent_files.notify(self.update)

    def update(self, files):
        for i, fname in enumerate(files):
            fname = fname.encode('utf-8')
            text = "&{} {}".format(i + 1, os.path.basename(fname))
            action = self._actions[i]
            action.setText(text)
            action.setData(fname)
            action.setVisible(True)



#    This is a component of AXIS, a front-end for emc
#    Copyright 2004, 2005, 2006 Jeff Epler <jepler@unpythonic.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import math
import gcode
import linuxcnc


class BaseCanon(object):
    def __init__(self):

        self.feedrate = 1
        self.dwell_time = 0

        self.seq_num = -1
        self.last_pos = (0,) * 9

        self.first_move = True
        self.in_arc = 0
        self.suppress = 0

        self.plane = 1
        self.arcdivision = 64

        # extents
        self.min_extents = [9e99, 9e99, 9e99]
        self.max_extents = [-9e99, -9e99, -9e99]
        self.min_extents_notool = [9e99, 9e99, 9e99]
        self.max_extents_notool = [-9e99, -9e99, -9e99]

        # tool length offsets
        self.tlo_x = 0.0
        self.tlo_y = 0.0
        self.tlo_z = 0.0
        self.tlo_a = 0.0
        self.tlo_b = 0.0
        self.tlo_c = 0.0
        self.tlo_u = 0.0
        self.tlo_v = 0.0
        self.tlo_w = 0.0

        self.tool_offsets = (0.0,) * 9

        # G92/G52 offsets
        self.g92_offset_x = 0.0
        self.g92_offset_y = 0.0
        self.g92_offset_z = 0.0
        self.g92_offset_a = 0.0
        self.g92_offset_b = 0.0
        self.g92_offset_c = 0.0
        self.g92_offset_u = 0.0
        self.g92_offset_v = 0.0
        self.g92_offset_w = 0.0
        self.g5x_index = 1

        # g5x offsets
        self.g5x_offset_x = 0.0
        self.g5x_offset_y = 0.0
        self.g5x_offset_z = 0.0
        self.g5x_offset_a = 0.0
        self.g5x_offset_b = 0.0
        self.g5x_offset_c = 0.0
        self.g5x_offset_u = 0.0
        self.g5x_offset_v = 0.0
        self.g5x_offset_w = 0.0

        # XY rotation (degrees)
        self.rotation_xy = 0
        self.rotation_cos = 1
        self.rotation_sin = 0

    def add_path_point(self, line_type, start_point, end_point):
        pass

    def comment(self, msg):
        pass

    def message(self, msg):
        pass

    def check_abort(self):
        pass

    def next_line(self, st):
        # state attributes
        # 'block', 'cutter_side', 'distance_mode', 'feed_mode', 'feed_rate',
        # 'flood', 'gcodes', 'mcodes', 'mist', 'motion_mode', 'origin', 'units',
        # 'overrides', 'path_mode', 'plane', 'retract_mode', 'sequence_number',
        # 'speed', 'spindle', 'stopping', 'tool_length_offset', 'toolchange',
        self.state = st
        self.seq_num = self.state.sequence_number

    def calc_extents(self):
        self.min_extents, self.max_extents, self.min_extents_notool, \
        self.max_extents_notool = gcode.calc_extents(self.arcfeed, self.feed, self.traverse)

    def rotate_and_translate(self, x, y, z, a, b, c, u, v, w):
        x += self.g92_offset_x
        y += self.g92_offset_y
        z += self.g92_offset_z
        a += self.g92_offset_a
        b += self.g92_offset_b
        c += self.g92_offset_c
        u += self.g92_offset_u
        v += self.g92_offset_v
        w += self.g92_offset_w

        if self.rotation_xy:
            rotx = x * self.rotation_cos - y * self.rotation_sin
            roty = x * self.rotation_sin + y * self.rotation_cos
            x, y = rotx, roty

        x += self.g5x_offset_x
        y += self.g5x_offset_y
        z += self.g5x_offset_z
        a += self.g5x_offset_a
        b += self.g5x_offset_b
        c += self.g5x_offset_c
        u += self.g5x_offset_u
        v += self.g5x_offset_v
        w += self.g5x_offset_w

        return [x, y, z, a, b, c, u, v, w]

    def set_g5x_offset(self, index, x, y, z, a, b, c, u, v, w):
        self.g5x_index = index
        self.g5x_offset_x = x
        self.g5x_offset_y = y
        self.g5x_offset_z = z
        self.g5x_offset_a = a
        self.g5x_offset_b = b
        self.g5x_offset_c = c
        self.g5x_offset_u = u
        self.g5x_offset_v = v
        self.g5x_offset_w = w

    def set_g92_offset(self, x, y, z, a, b, c, u, v, w):
        self.g92_offset_x = x
        self.g92_offset_y = y
        self.g92_offset_z = z
        self.g92_offset_a = a
        self.g92_offset_b = b
        self.g92_offset_c = c
        self.g92_offset_u = u
        self.g92_offset_v = v
        self.g92_offset_w = w

    def set_xy_rotation(self, rotation):
        self.rotation_xy = rotation
        theta = math.radians(rotation)
        self.rotation_cos = math.cos(theta)
        self.rotation_sin = math.sin(theta)

    def tool_offset(self, xo, yo, zo, ao, bo, co, uo, vo, wo):
        self.first_move = True
        x, y, z, a, b, c, u, v, w = self.last_pos

        self.last_pos = (
            x - xo + self.tlo_x, y - yo + self.tlo_y, z - zo + self.tlo_z,
            a - ao + self.tlo_a, b - bo + self.tlo_b, c - bo + self.tlo_b,
            u - uo + self.tlo_u, v - vo + self.tlo_v, w - wo + self.tlo_w)

        self.tlo_x = xo
        self.tlo_y = yo
        self.tlo_z = zo
        self.tlo_a = ao
        self.tlo_b = bo
        self.tlo_c = co
        self.tlo_u = uo
        self.tlo_v = vo
        self.tlo_w = wo

        self.tool_offsets = xo, yo, zo, ao, bo, co, uo, vo, wo

    def set_spindle_rate(self, speed):
        pass

    def set_feed_rate(self, feed_rate):
        self.feedrate = feed_rate / 60.

    def select_plane(self, plane):
        pass

    def change_tool(self, pocket):
        self.first_move = True

    def get_tool(self, pocket):
        return -1, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0

    def straight_traverse(self, x, y, z, a, b, c, u, v, w):
        if self.suppress > 0:
            return

        pos = self.rotate_and_translate(x, y, z, a, b, c, u, v, w)
        if not self.first_move:
            self.add_path_point('traverse', self.last_pos, pos)

        self.last_pos = pos

    def rigid_tap(self, x, y, z):
        if self.suppress > 0:
            return

        self.first_move = False
        pos = self.rotate_and_translate(x, y, z, 0, 0, 0, 0, 0, 0)[:3]
        pos += self.last_pos[3:]

        self.add_path_point('feed', self.last_pos, pos)

    def set_plane(self, plane):
        self.plane = plane

    def arc_feed(self, end_x, end_y, center_x, center_y, rot, end_z, a, b, c, u, v, w):

        if self.suppress > 0:
            return

        self.first_move = False
        self.in_arc = True
        try:
            self.lo = tuple(self.last_pos)
            segs = gcode.arc_to_segments(self, end_x, end_y, center_x, center_y,
                                         rot, end_z, a, b, c, u, v, w, self.arcdivision)
            self.straight_arcsegments(segs)
        finally:
            self.in_arc = False

    def straight_arcsegments(self, segs):
        self.first_move = False
        last_pos = self.last_pos
        for pos in segs:
            self.add_path_point('arcfeed', last_pos, pos)
            last_pos = pos
        self.last_pos = last_pos

    def straight_feed(self, x, y, z, a, b, c, u, v, w):
        if self.suppress > 0:
            return

        self.first_move = False
        pos = self.rotate_and_translate(x, y, z, a, b, c, u, v, w)

        self.add_path_point('feed', self.last_pos, pos)
        self.last_pos = pos

    straight_probe = straight_feed

    def user_defined_function(self, i, p, q):
        if self.suppress > 0:
            return

        self.add_path_point('user', self.last_pos, self.last_pos)

    def dwell(self, arg):
        if self.suppress > 0:
            return

        self.dwell_time += arg
        self.add_path_point('dwell', self.last_pos, self.last_pos)

    def get_external_angular_units(self):
        return 1.0

    def get_external_length_units(self):
        return 1.0

    def get_axis_mask(self):
        return 0

    def get_block_delete(self):
        return 0


class StatCanon(BaseCanon):
    def __init__(self, geometry='XYZ', random=False, stat=None):
        super(StatCanon, self).__init__()

        self.geometry = geometry
        self.random = random

        self.stat = stat or linuxcnc.stat()
        self.stat.poll()

        self.tools = list(self.stat.tool_table)

    def change_tool(self, pocket):
        if self.random:
            self.tools[0] = self.tools[pocket]
            self.tools[pocket] = self.tools[0]
        elif pocket == 0:
            self.tools[0] = (-1, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
                             0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0)
        else:
            self.tools[0] = self.tools[pocket]

    def get_tool(self, pocket):
        if pocket >= 0 and pocket < len(self.tools):
            return tuple(self.tools[pocket])
        return -1, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0

    def get_external_angular_units(self):
        return self.stat.angular_units or 1.0

    def get_external_length_units(self):
        return self.stat.linear_units or 1.0

    def get_axis_mask(self):
        return self.stat.axis_mask

    def get_block_delete(self):
        return self.stat.block_delete


class PrintCanon(BaseCanon):
    def set_g5x_offset(self, *args):
        print "set_g5x_offset", args

    def set_g92_offset(self, *args):
        print "set_g92_offset", args

    def next_line(self, state):
        print "next_line", state.sequence_number
        self.state = state

    def set_plane(self, plane):
        print "set plane", plane

    def set_feed_rate(self, arg):
        print "set feed rate", arg

    def comment(self, arg):
        print "#", arg

    def straight_traverse(self, *args):
        print "straight_traverse %.4g %.4g %.4g  %.4g %.4g %.4g   %.4g %.4g %.4g" % args

    def straight_feed(self, *args):
        print "straight_feed %.4g %.4g %.4g  %.4g %.4g %.4g  %.4g %.4g %.4g" % args

    def dwell(self, arg):
        if arg < .1:
            print "dwell %f ms" % (1000 * arg)
        else:
            print "dwell %f seconds" % arg

    def arc_feed(self, *args):
        print "arc_feed %.4g %.4g  %.4g %.4g %.4g  %.4g  %.4g %.4g %.4g" % args

    def get_axis_mask(self):
        return 7  # XYZ



"""QPin"""

import signal
import _hal
import hal

from qtpy.QtCore import QObject, Signal, QTimer


class QPin(QObject):
    valueChanged = Signal(object)
    valueIncreased = Signal(object)
    valueDecreased = Signal(object)

    def __init__(self, comp, name, typ, dir):
        super(QPin, self).__init__()

        self._pin = _hal.component.newpin(comp, name, typ, dir)
        self._val = self._pin.get()

        self.startTimer(100)

    def timerEvent(self, timer):
        tmp = self._pin.get()
        if tmp != self._val:
            self.valueChanged.emit(tmp)
            if tmp > self._val:
                self.valueIncreased.emit(tmp)
            else:
                self.valueDecreased.emit(tmp)
            self._val = tmp

    @property
    def value(self):
        return self._pin.get()

    @value.setter
    def value(self, val):
        self._pin.set(val)


class QComponent(QObject):
    def __init__(self, comp_name):
        super(QComponent, self).__init__()

        signal.signal(signal.SIGTERM, self.signal_handler)
        signal.signal(signal.SIGINT, self.signal_handler)

        self.type_map = {
            'float': hal.HAL_FLOAT,
            's32': hal.HAL_S32,
            'u32': hal.HAL_U32,
            'bit': hal.HAL_BIT
        }

        self.dir_map = {
            'in': hal.HAL_IN,
            'out': hal.HAL_OUT,
            'io': hal.HAL_IO
        }

        self._comp = _hal.component(comp_name)
        self._pins = {}

    def newPin(self, name, type, direction):

        pin_type = self.type_map.get(type.lower())
        pin_dir = self.dir_map.get(direction.lower())

        pin = QPin(self._comp, name, pin_type, pin_dir)
        self._pins[name] = pin

    def getPin(self, *a, **kw):
        return QPin(self.getpin(*a, **kw))

    def ready(self):
        self._comp.ready()

    def exit(self, *a, **kw):
        print("exit")
        return self._comp.exit(*a, **kw)

    def signal_handler(self, signal, frame):
        self.exit()

    def __getitem__(self, item):
        return self._pins[item]


def main():
    from qtpy.QtWidgets import QApplication

    app = QApplication([])

    def printChange(new_val):
        print("Value Changed", new_val)

    c = QComponent('test')
    c.newPin('input', "s32", "in")
    c.newPin('float_in', "s32", "in")
    c['input'].valueChanged.connect(printChange)
    c.ready()

    app.exec_()


if __name__ == "__main__":
    main()
